@@ -10,10 +10,17 @@ module Asciidoctor
     TAB = %(\t)
     WHITESPACE = %(#{LF}#{TAB} )
     ET = ' ' * 8
-    ESC = %(\u001b)      # troff leader marker
+    ESC = %(\u001b) # troff leader marker
     ESC_BS = %(#{ESC}\\) # escaped backslash (indicates troff formatting sequence)
     ESC_FS = %(#{ESC}.)  # escaped full stop (indicates troff macro)
 
+    LiteralBackslashRx = /(?:\A|[^#{ESC}])\\/
+    LeadingPeriodRx = /^\./
+    EscapedMacroRx = /^(?:#{ESC}\\c\n)?#{ESC}\.((?:URL|MTO) ".*?" ".*?" )( |[^\s]*)(.*?)(?: *#{ESC}\\c)?$/
+    MockBoundaryRx = /<\/?BOUNDARY>/
+    EmDashCharRefRx = /&#8212;(?:&#8203;)?/
+    EllipsisCharRefRx = /&#8230;(?:&#8203;)?/
+
     # Converts HTML entity references back to their original form, escapes
     # special man characters and strips trailing whitespace.
     #
@@ -26,12 +33,10 @@ module Asciidoctor
     #        * :append_newline a Boolean that indicates whether to append an endline to the result (default: false)
     def manify str, opts = {}
       str = ((opts.fetch :preserve_space, true) ? (str.gsub TAB, ET) : (str.tr_s WHITESPACE, ' ')).
-        gsub(/(?:\A|[^#{ESC}])\\/, '\&(rs'). # literal backslash (not a troff escape sequence)
-        gsub(/^\./, '\\\&.').     # leading . is used in troff for macro call or other formatting; replace with \&.
+        gsub(LiteralBackslashRx, '\&(rs'). # literal backslash (not a troff escape sequence)
+        gsub(LeadingPeriodRx, '\\\&.'). # leading . is used in troff for macro call or other formatting; replace with \&.
         # drop orphaned \c escape lines, unescape troff macro, quote adjacent character, isolate macro line
-        gsub(/^(?:#{ESC}\\c\n)?#{ESC}\.((?:URL|MTO) ".*?" ".*?" )( |[^\s]*)(.*?)(?: *#{ESC}\\c)?$/) {
-          (rest = $3.lstrip).empty? ? %(.#$1"#$2") : %(.#$1"#$2"#{LF}#{rest})
-        }.
+        gsub(EscapedMacroRx) { (rest = $3.lstrip).empty? ? %(.#$1"#$2") : %(.#$1"#$2"#{LF}#{rest}) }.
         gsub('-', '\-').
         gsub('&lt;', '<').
         gsub('&gt;', '>').
@@ -41,26 +46,28 @@ module Asciidoctor
         gsub('&#8482;', '\(tm').  # trademark sign
         gsub('&#8201;', ' ').     # thin space
         gsub('&#8211;', '\(en').  # en dash
-        gsub(/&#8212(?:;&#8203;)?/, '\(em'). # em dash
+        gsub(EmDashCharRefRx, '\(em'). # em dash
         gsub('&#8216;', '\(oq').  # left single quotation mark
         gsub('&#8217;', '\(cq').  # right single quotation mark
         gsub('&#8220;', '\(lq').  # left double quotation mark
         gsub('&#8221;', '\(rq').  # right double quotation mark
-        gsub(/&#8230;(?:&#8203;)?/, '...'). # horizontal ellipsis
+        gsub(EllipsisCharRefRx, '...'). # horizontal ellipsis
         gsub('&#8592;', '\(<-').  # leftwards arrow
         gsub('&#8594;', '\(->').  # rightwards arrow
         gsub('&#8656;', '\(lA').  # leftwards double arrow
         gsub('&#8658;', '\(rA').  # rightwards double arrow
         gsub('&#8203;', '\:').    # zero width space
+        gsub('&amp;','&').        # literal ampersand (NOTE must take place after any other replacement that includes &)
         gsub('\'', '\(aq').       # apostrophe-quote
-        gsub(/<\/?BOUNDARY>/, '').# artificial boundary
-        gsub(ESC_BS, '\\').       # unescape troff backslash (NOTE update if more escaped are added)
+        gsub(MockBoundaryRx, ''). # mock boundary
+        gsub(ESC_BS, '\\').       # unescape troff backslash (NOTE update if more escapes are added)
+        gsub(ESC_FS, '.').        # unescape full stop in troff commands (NOTE must take place after gsub(LeadingPeriodRx))
         rstrip                    # strip trailing space
       opts[:append_newline] ? %(#{str}#{LF}) : str
     end
 
     def skip_with_warning node, name = nil
-      warn %(asciidoctor: WARNING: converter missing for #{name || node.node_name} node in manpage backend)
+      logger.warn %(converter missing for #{name || node.node_name} node in manpage backend)
       nil
     end
 
@@ -75,7 +82,7 @@ module Asciidoctor
       # NOTE the first line enables the table (tbl) preprocessor, necessary for non-Linux systems
       result = [%('\\" t
 .\\"     Title: #{mantitle}
-.\\"    Author: #{(node.attr? 'authors') ? (node.attr 'authors') : '[see the "AUTHORS" section]'}
+.\\"    Author: #{(node.attr? 'authors') ? (node.attr 'authors') : '[see the "AUTHOR(S)" section]'}
 .\\" Generator: Asciidoctor #{node.attr 'asciidoctor-version'})]
       result << %(.\\"      Date: #{docdate}) if docdate
       result << %(.\\"    Manual: #{(manual = node.attr 'manmanual') || '\ \&'}
@@ -99,22 +106,33 @@ module Asciidoctor
       # define URL macro for portability
       # see http://web.archive.org/web/20060102165607/http://people.debian.org/~branden/talks/wtfm/wtfm.pdf
       #
-      # Use: .URL "http://www.debian.org" "Debian" "."
+      # Usage
+      #
+      # .URL "http://www.debian.org" "Debian" "."
       #
       # * First argument: the URL
       # * Second argument: text to be hyperlinked
-      # * Third (optional) argument: text that needs to immediately trail
-      #   the hyperlink without intervening whitespace
+      # * Third (optional) argument: text that needs to immediately trail the hyperlink without intervening whitespace
       result << '.de URL
-\\\\$2 \(laURL: \\\\$1 \(ra\\\\$3
+\\fI\\\\$2\\fP <\\\\$1>\\\\$3
 ..
-.if \n[.g] .mso www.tmac'
-      result << %(.LINKSTYLE #{node.attr 'man-linkstyle', 'blue R < >'})
+.als MTO URL
+.if \n[.g] \{\
+.  mso www.tmac
+.  am URL
+.    ad l
+.  .
+.  am MTO
+.    ad l
+.  .'
+      result << %(.  LINKSTYLE #{node.attr 'man-linkstyle', 'blue R < >'})
+      result << '.\}'
 
       unless node.noheader
         if node.attr? 'manpurpose'
-          result << %(.SH "#{node.attr 'manname-title'}"
-#{manify mantitle} \\- #{manify node.attr 'manpurpose'})
+          mannames = node.attr 'mannames', [manname]
+          result << %(.SH "#{(node.attr 'manname-title').upcase}"
+#{mannames.map {|n| manify n } * ', '} \\- #{manify node.attr 'manpurpose'})
         end
       end
 
@@ -126,14 +144,19 @@ module Asciidoctor
         result.concat(node.footnotes.map {|fn| %(#{fn.index}. #{fn.text}) })
       end
 
-      # FIXME detect single author and use appropriate heading; itemize the authors if multiple
-      if node.attr? 'authors'
-        result << %(.SH "AUTHOR(S)"
+      # FIXME we really need an API that returns the authors as an array
+      if (num_authors = (node.attr 'authorcount') || 0) > 0
+        if num_authors == 1
+          result << %(.SH "AUTHOR"
 .sp
-\\fB#{node.attr 'authors'}\\fP
-.RS 4
-Author(s).
-.RE)
+#{node.attr 'author'})
+        else
+          result << '.SH "AUTHORS"'
+          (1.upto num_authors).each do |i|
+            result << %(.sp
+#{node.attr "author_#{i}"})
+          end
+        end
       end
 
       result * LF
@@ -154,13 +177,10 @@ Author(s).
     end
 
     def section node
-      slevel = node.level
-      # QUESTION should the check for slevel be done in section?
-      slevel = 1 if slevel == 0 && node.special
       result = []
-      if slevel > 1
+      if node.level > 1
         macro = 'SS'
-        # QUESTION why captioned title? why not for slevel == 1?
+        # QUESTION why captioned title? why not when level == 1?
         stitle = node.captioned_title
       else
         macro = 'SH'
@@ -173,16 +193,14 @@ Author(s).
 
     def admonition node
       result = []
-      result << %(.if n \\{\\
-.sp
-.\\}
+      result << %(.if n .sp
 .RS 4
 .it 1 an-trap
 .nr an-no-space-flag 1
 .nr an-break-flag 1
 .br
 .ps +1
-.B #{node.caption}#{node.title? ? "\\fP #{manify node.title}" : nil}
+.B #{node.attr 'textlabel'}#{node.title? ? "\\fP: #{manify node.title}" : ''}
 .ps -1
 .br
 #{resolve_content node}
@@ -191,7 +209,7 @@ Author(s).
       result * LF
     end
 
-    alias :audio :skip_with_warning
+    alias audio skip_with_warning
 
     def colist node
       result = []
@@ -202,30 +220,34 @@ Author(s).
 tab(:);
 r lw(\n(.lu*75u/100u).'
 
-      node.items.each_with_index do |item, index|
-        result << %(\\fB(#{index + 1})\\fP\\h'-2n':T{
-#{manify item.text}
-T})
+      num = 0
+      node.items.each do |item|
+        result << %(\\fB(#{num += 1})\\fP\\h'-2n':T{)
+        result << (manify item.text)
+        result << item.content if item.blocks?
+        result << 'T}'
       end
       result << '.TE'
       result * LF
     end
 
-    # TODO implement title for dlist
     # TODO implement horizontal (if it makes sense)
     def dlist node
       result = []
+      result << %(.sp
+.B #{manify node.title}
+.br) if node.title?
       counter = 0
       node.items.each do |terms, dd|
         counter += 1
         case node.style
         when 'qanda'
           result << %(.sp
-#{counter}. #{manify([*terms].map {|dt| dt.text }.join ' ')}
+#{counter}. #{manify([*terms].map {|dt| dt.text } * ' ')}
 .RS 4)
         else
           result << %(.sp
-#{manify([*terms].map {|dt| dt.text }.join ', ')}
+#{manify([*terms].map {|dt| dt.text } * ', ')}
 .RS 4)
         end
         if dd
@@ -252,7 +274,7 @@ T})
       %(.SS "#{manify node.title}")
     end
 
-    alias :image :skip_with_warning
+    alias image skip_with_warning
 
     def listing node
       result = []
@@ -260,15 +282,11 @@ T})
 .B #{manify node.captioned_title}
 .br) if node.title?
       result << %(.sp
-.if n \\{\\
-.RS 4
-.\\}
+.if n .RS 4
 .nf
 #{manify node.content}
 .fi
-.if n \\{\\
-.RE
-.\\})
+.if n .RE)
       result * LF
     end
 
@@ -278,15 +296,11 @@ T})
 .B #{manify node.title}
 .br) if node.title?
       result << %(.sp
-.if n \\{\\
-.RS 4
-.\\}
+.if n .RS 4
 .nf
 #{manify node.content}
 .fi
-.if n \\{\\
-.RE
-.\\})
+.if n .RE)
       result * LF
     end
 
@@ -303,8 +317,8 @@ T})
 \\h'-04' #{idx + 1}.\\h'+01'\\c
 .\\}
 .el \\{\\
-.sp -1
-.IP " #{idx + 1}." 4.2
+.  sp -1
+.  IP " #{idx + 1}." 4.2
 .\\}
 #{manify item.text})
         result << item.content if item.blocks?
@@ -323,7 +337,7 @@ T})
     end
 
     # TODO use Page Control https://www.gnu.org/software/groff/manual/html_node/Page-Control.html#Page-Control
-    alias :page_break :skip
+    alias page_break skip
 
     def paragraph node
       if node.title?
@@ -337,7 +351,7 @@ T})
       end
     end
 
-    alias :preamble :content
+    alias preamble content
 
     def quote node
       result = []
@@ -368,12 +382,12 @@ T})
       result * LF
     end
 
-    alias :sidebar :skip_with_warning
+    alias sidebar skip_with_warning
 
     def stem node
       title_element = node.title? ? %(.sp
 .B #{manify node.title}
-.br) : nil
+.br) : ''
       open, close = BLOCK_MATH_DELIMITERS[node.style.to_sym]
 
       unless ((equation = node.content).start_with? open) && (equation.end_with? close)
@@ -397,15 +411,16 @@ T})
 .nr an-no-space-flag 1
 .nr an-break-flag 1
 .br
-.B #{manify node.captioned_title})
+.B #{manify node.captioned_title}
+)
       end
       result << '.TS
 allbox tab(:);'
       row_header = []
       row_text = []
       row_index = 0
-      [:head, :body, :foot].each do |tsec|
-        node.rows[tsec].each do |row|
+      node.rows.by_section.each do |tsec, rows|
+        rows.each do |row|
           row_header[row_index] ||= []
           row_text[row_index] ||= []
           # result << LF
@@ -425,19 +440,17 @@ allbox tab(:);'
               row_text[row_index] << %(T{#{LF}.sp#{LF}T}:)
             end
             row_text[row_index] << %(T{#{LF}.sp#{LF})
-            cell_halign = (cell.attr 'halign', 'left')[0..0]
+            cell_halign = (cell.attr 'halign', 'left').chr
             if tsec == :head
-              if row_header[row_index].empty? ||
-                 row_header[row_index][cell_index].empty?
+              if row_header[row_index].empty? || row_header[row_index][cell_index].empty?
                 row_header[row_index][cell_index] << %(#{cell_halign}tB)
               else
                 row_header[row_index][cell_index + 1] ||= []
                 row_header[row_index][cell_index + 1] << %(#{cell_halign}tB)
               end
-              row_text[row_index] << %(#{cell.text}#{LF})
+              row_text[row_index] << %(#{manify cell.text}#{LF})
             elsif tsec == :body
-              if row_header[row_index].empty? ||
-                 row_header[row_index][cell_index].empty?
+              if row_header[row_index].empty? || row_header[row_index][cell_index].empty?
                 row_header[row_index][cell_index] << %(#{cell_halign}t)
               else
                 row_header[row_index][cell_index + 1] ||= []
@@ -446,26 +459,26 @@ allbox tab(:);'
               case cell.style
               when :asciidoc
                 cell_content = cell.content
-              when :verse, :literal
-                cell_content = cell.text
+              when :literal
+                cell_content = %(.nf#{LF}#{manify cell.text}#{LF}.fi)
+              when :verse
+                cell_content = %(.nf#{LF}#{manify cell.text}#{LF}.fi)
               else
-                cell_content = cell.content.join
+                cell_content = manify cell.content.join
               end
               row_text[row_index] << %(#{cell_content}#{LF})
             elsif tsec == :foot
-              if row_header[row_index].empty? ||
-                 row_header[row_index][cell_index].empty?
+              if row_header[row_index].empty? || row_header[row_index][cell_index].empty?
                 row_header[row_index][cell_index] << %(#{cell_halign}tB)
               else
                 row_header[row_index][cell_index + 1] ||= []
                 row_header[row_index][cell_index + 1] << %(#{cell_halign}tB)
               end
-              row_text[row_index] << %(#{cell.text}#{LF})
+              row_text[row_index] << %(#{manify cell.text}#{LF})
             end
             if cell.colspan && cell.colspan > 1
               (cell.colspan - 1).times do |i|
-                if row_header[row_index].empty? ||
-                   row_header[row_index][cell_index].empty?
+                if row_header[row_index].empty? || row_header[row_index][cell_index].empty?
                   row_header[row_index][cell_index + i] << 'st'
                 else
                   row_header[row_index][cell_index + 1 + i] ||= []
@@ -476,8 +489,7 @@ allbox tab(:);'
             if cell.rowspan && cell.rowspan > 1
               (cell.rowspan - 1).times do |i|
                 row_header[row_index + 1 + i] ||= []
-                if row_header[row_index + 1 + i].empty? ||
-                   row_header[row_index + 1 + i][cell_index].empty?
+                if row_header[row_index + 1 + i].empty? || row_header[row_index + 1 + i][cell_index].empty?
                   row_header[row_index + 1 + i][cell_index] ||= []
                   row_header[row_index + 1 + i][cell_index] << '^t'
                 else
@@ -493,19 +505,19 @@ allbox tab(:);'
             end
           end
           row_index += 1
-        end
+        end unless rows.empty?
       end
 
       #row_header.each do |row|
       #  result << LF
       #  row.each_with_index do |cell, i|
-      #    result << (cell.join ' ')
+      #    result << (cell * ' ')
       #    result << ' ' if row.size > i + 1
       #  end
       #end
       # FIXME temporary fix to get basic table to display
       result << LF
-      result << row_header.first.map {|r| 'lt'}.join(' ')
+      result << row_header[0].map { 'lt' } * ' '
 
       result << %(.#{LF})
       row_text.each do |row|
@@ -521,7 +533,7 @@ allbox tab(:);'
 \l\'\n(.lu*25u/100u\(ap\''
     end
 
-    alias :toc :skip
+    alias toc skip
 
     def ulist node
       result = []
@@ -535,8 +547,8 @@ allbox tab(:);'
 \\h'-04'\\(bu\\h'+03'\\c
 .\\}
 .el \\{\\
-.sp -1
-.IP \\(bu 2.3
+.  sp -1
+.  IP \\(bu 2.3
 .\\}
 #{manify item.text}]
         result << item.content if item.blocks?
@@ -571,42 +583,47 @@ allbox tab(:);'
     end
 
     def video node
-      start_param = (node.attr? 'start', nil, false) ? %(&start=#{node.attr 'start'}) : nil
-      end_param = (node.attr? 'end', nil, false) ? %(&end=#{node.attr 'end'}) : nil
-      %(.sp
-#{manify node.captioned_title} (video) <#{node.media_uri(node.attr 'target')}#{start_param}#{end_param}>)
+      start_param = (node.attr? 'start', nil, false) ? %(&start=#{node.attr 'start'}) : ''
+      end_param = (node.attr? 'end', nil, false) ? %(&end=#{node.attr 'end'}) : ''
+      result = []
+      result << %(.sp
+.B #{manify node.title}
+.br) if node.title?
+      result << %(<#{node.media_uri(node.attr 'target')}#{start_param}#{end_param}> (video))
+      result * LF
     end
 
     def inline_anchor node
       target = node.target
       case node.type
       when :link
-        if (text = node.text) == target
-          text = nil
-        else
-          text = text.gsub '"', %[#{ESC_BS}(dq]
-        end
         if target.start_with? 'mailto:'
           macro = 'MTO'
-          target = target[7..-1].sub '@', %[#{ESC_BS}(at]
+          target = target.slice 7, target.length
         else
           macro = 'URL'
         end
+        if (text = node.text) == target
+          text = ''
+        else
+          text = text.gsub '"', %[#{ESC_BS}(dq]
+        end
+        target = target.sub '@', %[#{ESC_BS}(at] if macro == 'MTO'
         %(#{ESC_BS}c#{LF}#{ESC_FS}#{macro} "#{target}" "#{text}" )
       when :xref
         refid = (node.attr 'refid') || target
-        node.text || (node.document.references[:ids][refid] || %([#{refid}]))
+        node.text || (node.document.catalog[:ids][refid] || %([#{refid}]))
       when :ref, :bibref
-        # These are anchor points, which shouldn't be visual
+        # These are anchor points, which shouldn't be visible
         ''
       else
-        warn %(asciidoctor: WARNING: unknown anchor type: #{node.type.inspect})
+        logger.warn %(unknown anchor type: #{node.type.inspect})
+        nil
       end
     end
 
     def inline_break node
-      %(#{node.text}
-.br)
+      %(#{node.text}#{LF}#{ESC_FS}br)
     end
 
     def inline_button node
@@ -627,9 +644,7 @@ allbox tab(:);'
     end
 
     def inline_image node
-      # NOTE alt should always be set
-      alt_text = (node.attr? 'alt') ? (node.attr 'alt') : node.target
-      (node.attr? 'link') ? %([#{alt_text}] <#{node.attr 'link'}>) : %([#{alt_text}])
+      (node.attr? 'link') ? %([#{node.alt}] <#{node.attr 'link'}>) : %([#{node.alt}])
     end
 
     def inline_indexterm node
@@ -640,7 +655,7 @@ allbox tab(:);'
       if (keys = node.attr 'keys').size == 1
         keys[0]
       else
-        keys.join %(#{ESC_BS}0+#{ESC_BS}0)
+        keys * %(#{ESC_BS}0+#{ESC_BS}0)
       end
     end
 
@@ -648,7 +663,7 @@ allbox tab(:);'
       caret = %[#{ESC_BS}0#{ESC_BS}(fc#{ESC_BS}0]
       menu = node.attr 'menu'
       if !(submenus = node.attr 'submenus').empty?
-        submenu_path = submenus.map {|item| %(#{ESC_BS}fI#{item}#{ESC_BS}fP) }.join caret
+        submenu_path = submenus.map {|item| %(#{ESC_BS}fI#{item}#{ESC_BS}fP) } * caret
         %(#{ESC_BS}fI#{menu}#{ESC_BS}fP#{caret}#{submenu_path}#{caret}#{ESC_BS}fI#{node.attr 'menuitem'}#{ESC_BS}fP)
       elsif (menuitem = node.attr 'menuitem')
         %(#{ESC_BS}fI#{menu}#{caret}#{menuitem}#{ESC_BS}fP)
@@ -665,7 +680,7 @@ allbox tab(:);'
       when :strong
         %(#{ESC_BS}fB<BOUNDARY>#{node.text}</BOUNDARY>#{ESC_BS}fP)
       when :monospaced
-        %(#{ESC_BS}f[CR]<BOUNDARY>#{node.text}</BOUNDARY>#{ESC_BS}fP)
+        %[#{ESC_BS}f(CR<BOUNDARY>#{node.text}</BOUNDARY>#{ESC_BS}fP]
       when :single
         %[#{ESC_BS}(oq<BOUNDARY>#{node.text}</BOUNDARY>#{ESC_BS}(cq]
       when :double
@@ -678,5 +693,16 @@ allbox tab(:);'
     def resolve_content node
       node.content_model == :compound ? node.content : %(.sp#{LF}#{manify node.content})
     end
+
+    def write_alternate_pages mannames, manvolnum, target
+      if mannames && mannames.size > 1
+        mannames.shift
+        manvolext = %(.#{manvolnum})
+        dir, basename = ::File.split target
+        mannames.each do |manname|
+          ::IO.write ::File.join(dir, %(#{manname}#{manvolext})), %(.so #{basename})
+        end
+      end
+    end
   end
 end
