@@ -1,5 +1,7 @@
 # encoding: UTF-8
-ASCIIDOCTOR_PROJECT_DIR = File.dirname File.dirname(__FILE__)
+ASCIIDOCTOR_TEST_DIR = File.expand_path File.dirname __FILE__
+ASCIIDOCTOR_PROJECT_DIR = File.dirname ASCIIDOCTOR_TEST_DIR
+ASCIIDOCTOR_LIB_DIR = ENV['ASCIIDOCTOR_LIB_DIR'] || File.join(ASCIIDOCTOR_PROJECT_DIR, 'lib')
 Dir.chdir ASCIIDOCTOR_PROJECT_DIR
 
 if RUBY_VERSION < '1.9'
@@ -8,10 +10,11 @@ end
 
 require 'simplecov' if ENV['COVERAGE'] == 'true'
 
-require File.join(ASCIIDOCTOR_PROJECT_DIR, 'lib', 'asciidoctor')
+require File.join(ASCIIDOCTOR_LIB_DIR, 'asciidoctor')
 
 require 'socket'
 require 'nokogiri'
+require 'tempfile'
 require 'tmpdir'
 
 autoload :FileUtils, 'fileutils'
@@ -31,7 +34,7 @@ class Minitest::Test
   end
 
   def disk_root
-    "#{windows? ? File.expand_path(__FILE__).split('/').first : nil}/"
+    %(#{windows? ? ASCIIDOCTOR_PROJECT_DIR.split('/')[0] : ''}/)
   end
 
   def empty_document options = {}
@@ -60,37 +63,28 @@ class Minitest::Test
     fixture_path(name)
   end
 
-  def fixture_path(name)
-    File.join(File.expand_path(File.dirname(__FILE__)), 'fixtures', name)
+  def testdir
+    ASCIIDOCTOR_TEST_DIR
+  end
+
+  def fixturedir
+    File.join testdir, 'fixtures'
+  end
+
+  def fixture_path name
+    File.join fixturedir, name
   end
 
   def example_document(name, opts = {})
-    document_from_string File.read(sample_doc_path(name)), opts
-  end
-
-  def assert_difference(expression, difference = 1, message = nil, &block)
-    expressions = [expression]
-
-    exps = expressions.map { |e|
-      e.respond_to?(:call) ? e : lambda { eval(e, block.binding) }
-    }
-    before = exps.map { |e| e.call }
-
-    yield
-
-    expressions.zip(exps).each_with_index do |(code, e), i|
-      error  = "#{code.inspect} didn't change by #{difference}"
-      error  = "#{message}.\n#{error}" if message
-      assert_equal(before[i] + difference, e.call, error)
-    end
+    document_from_string IO.read(sample_doc_path(name)), opts
   end
 
   def xmlnodes_at_css(css, content, count = nil)
-    xmlnodes_at_path(:css, css, content)
+    xmlnodes_at_path(:css, css, content, count)
   end
 
   def xmlnodes_at_xpath(xpath, content, count = nil)
-    xmlnodes_at_path(:xpath, xpath, content)
+    xmlnodes_at_path(:xpath, xpath, content, count)
   end
 
   def xmlnodes_at_path(type, path, content, count = nil)
@@ -134,8 +128,8 @@ class Minitest::Test
       else
         assert true
       end
-    elsif (count && results.length != count)
-      flunk "#{type_name} #{path} yielded #{results.length} elements rather than #{count} for:\n#{content}"
+    elsif (count && results.size != count)
+      flunk "#{type_name} #{path} yielded #{results.size} elements rather than #{count} for:\n#{content}"
     elsif (count.nil? && results.empty?)
       flunk "#{type_name} #{path} not found in:\n#{content}"
     else
@@ -143,15 +137,44 @@ class Minitest::Test
     end
   end
 
+  def assert_message logger, severity, expected_message, kind = String, idx = nil
+    unless idx
+      assert_equal 1, logger.messages.size
+      idx = 0
+    end
+    message = logger.messages[idx]
+    assert_equal severity, message[:severity]
+    assert_kind_of kind, message[:message]
+    if kind == String
+      actual_message = message[:message]
+    else
+      refute_nil message[:message][:source_location]
+      actual_message = message[:message].inspect
+    end
+    if expected_message.start_with? '~'
+      assert_includes actual_message, expected_message[1..-1]
+    else
+      assert_equal expected_message, actual_message
+    end
+  end
+
+  def assert_messages logger, expected_messages
+    assert_equal expected_messages.size, logger.messages.size
+    expected_messages.each_with_index do |expected_message_details, idx|
+      severity, expected_message, kind = expected_message_details
+      assert_message logger, severity, expected_message, (kind || String), idx
+    end
+  end
+
   def xmldoc_from_string(content)
     if content.match(RE_XMLNS_ATTRIBUTE)
-      doc = Nokogiri::XML::Document.parse(content)
+      Nokogiri::XML::Document.parse(content)
     elsif !(doctype_match = content.match(RE_DOCTYPE))
-      doc = Nokogiri::HTML::DocumentFragment.parse(content)
+      Nokogiri::HTML::DocumentFragment.parse(content)
     elsif doctype_match[1].start_with? 'html'
-      doc = Nokogiri::HTML::Document.parse(content)
+      Nokogiri::HTML::Document.parse(content)
     else
-      doc = Nokogiri::XML::Document.parse(content)
+      Nokogiri::XML::Document.parse(content)
     end
   end
 
@@ -173,10 +196,10 @@ class Minitest::Test
   def render_string(src, opts = {})
     keep_namespaces = opts.delete(:keep_namespaces)
     if keep_namespaces
-      document_from_string(src, opts).render
+      document_from_string(src, opts).convert
     else
       # this is required because nokogiri is ignorant
-      result = document_from_string(src, opts).render
+      result = document_from_string(src, opts).convert
       result = result.sub(RE_XMLNS_ATTRIBUTE, '') if result
       result
     end
@@ -184,12 +207,17 @@ class Minitest::Test
 
   def render_embedded_string(src, opts = {})
     opts[:header_footer] = false
-    document_from_string(src, opts).render
+    document_from_string(src, opts).convert
   end
 
-  def parse_header_metadata(source)
-    reader = Asciidoctor::Reader.new source.split ::Asciidoctor::EOL
-    [::Asciidoctor::Parser.parse_header_metadata(reader), reader]
+  def render_inline_string(src, opts = {})
+    opts[:doctype] = :inline
+    document_from_string(src, opts).convert
+  end
+
+  def parse_header_metadata(source, doc = nil)
+    reader = Asciidoctor::Reader.new source.split ::Asciidoctor::LF
+    [::Asciidoctor::Parser.parse_header_metadata(reader, doc), reader]
   end
 
   def assign_default_test_options(opts)
@@ -206,34 +234,31 @@ class Minitest::Test
     end
     if (template_dir = ENV['TEMPLATE_DIR'])
       opts[:template_dir] = template_dir unless opts.has_key? :template_dir
-      #opts[:template_dir] = File.join(File.dirname(__FILE__), '..', '..', 'asciidoctor-backends', 'erb') unless opts.has_key? :template_dir
     end
     nil
   end
 
-  # Expand the character for an entity such as &#8212; into a glyph
-  # so it can be used to match in an XPath expression
+  # Decode the numeric character reference, such as 8212, to a Unicode glyph
+  # so it may be used in an XPath expression.
   #
   # Examples
   #
-  #   expand_entity 60
+  #   decode_char 60
   #   # => "<"
   #
-  # Returns the String entity expanded to its equivalent UTF-8 glyph
-  def expand_entity(number)
-    [number].pack('U*')
+  # Returns the decoded String that corresponds to the numeric character reference
+  def decode_char number
+    [number].pack 'U1'
   end
-  alias :entity :expand_entity
 
   def invoke_cli_with_filenames(argv = [], filenames = [], &block)
-
     filepaths = Array.new
 
-    filenames.each { |filename|
-      if filenames.nil?|| ::Pathname.new(filename).absolute?
-        filepaths.push(filename)
+    filenames.each {|filename|
+      if filenames.nil? || ::Pathname.new(filename).absolute?
+        filepaths << filename
       else
-        filepaths.push(File.join(File.dirname(__FILE__), 'fixtures', filename))
+        filepaths << (fixture_path filename)
       end
     }
 
@@ -251,7 +276,7 @@ class Minitest::Test
     if filename.nil? || filename == '-' || ::Pathname.new(filename).absolute?
       filepath = filename
     else
-      filepath = File.join(File.dirname(__FILE__), 'fixtures', filename)
+      filepath = fixture_path filename
     end
     invoker = Asciidoctor::Cli::Invoker.new(argv + [filepath])
     if buffers
@@ -264,12 +289,14 @@ class Minitest::Test
   def redirect_streams
     old_stdout, $stdout = $stdout, (tmp_stdout = ::StringIO.new)
     old_stderr, $stderr = $stderr, (tmp_stderr = ::StringIO.new)
-    begin
-      yield tmp_stdout, tmp_stderr
-    ensure
-      $stdout = old_stdout
-      $stderr = old_stderr
-    end
+    old_logger = Asciidoctor::LoggerManager.logger
+    old_logger_level = old_logger.level
+    new_logger = (Asciidoctor::LoggerManager.logger = Asciidoctor::Logger.new $stderr)
+    new_logger.level = old_logger_level
+    yield tmp_stdout, tmp_stderr
+  ensure
+    $stdout, $stderr = old_stdout, old_stderr
+    Asciidoctor::LoggerManager.logger = old_logger
   end
 
   def resolve_localhost
@@ -277,9 +304,29 @@ class Minitest::Test
         Socket.ip_address_list.find {|addr| addr.ipv4? }.ip_address
   end
 
+  def using_memory_logger
+    old_logger = Asciidoctor::LoggerManager.logger
+    memory_logger = Asciidoctor::MemoryLogger.new
+    begin
+      Asciidoctor::LoggerManager.logger = memory_logger
+      yield memory_logger
+    ensure
+      Asciidoctor::LoggerManager.logger = old_logger
+    end
+  end
+
+  def in_verbose_mode
+    begin
+      old_verbose, $VERBOSE = $VERBOSE, true
+      yield
+    ensure
+      $VERBOSE = old_verbose
+    end
+  end
+
   def using_test_webserver host = resolve_localhost, port = 9876
     server = TCPServer.new host, port
-    base_dir = File.expand_path File.dirname __FILE__
+    base_dir = testdir
     t = Thread.new do
       while (session = server.accept)
         request = session.gets
